@@ -2,7 +2,7 @@
 
 """ Browse & play embedded tracks from Angry Metal Guy music reviews. """
 
-__version__ = "2020.04.25.0"
+__version__ = "2020.07.16.0"
 __author__ = "desbma"
 __license__ = "GPLv3"
 
@@ -665,9 +665,6 @@ def cl_main():
     # fetch review & play
     review_page = fetch_page(review.url, http_cache=http_cache)
     header = REVIEW_HEADER_SELECTOR(review_page)[0]
-    for child in header:
-      if child.tag == "i":
-        header.remove(child)
     date_published = lxml.etree.tostring(header, encoding="unicode", method="text").strip()
     date_published = REVIEW_HEADER_DATE_REGEX.search(date_published).group(1)
     with date_locale_neutral():
@@ -689,8 +686,8 @@ def cl_main():
           while c not in frozenset("pdrsq"):
             c = input("[P]lay / [D]ownload / Go to [R]eview / [S]kip to next track / Exit [Q] ? ").lower()
           if c == "p":
-            known_reviews.setLastPlayed(review.url)
             play(review, track_urls, merge_with_picture=audio_only)
+            known_reviews.setLastPlayed(review.url)
             input_loop = False
           elif c == "d":
             download_audio(review, date_published, track_urls, max_cover_size=args.max_embedded_cover_size)
@@ -703,13 +700,13 @@ def cl_main():
             input_loop = False
             track_loop = False
       else:
-        known_reviews.setLastPlayed(review.url)
         if (((args.mode in (PlayerMode.MANUAL, PlayerMode.RADIO)) and
                 (action is menu.AmgMenu.UserAction.DOWNLOAD_AUDIO)) or
                 (args.mode is PlayerMode.DISCOVER_DOWNLOAD)):
           download_audio(review, date_published, track_urls, max_cover_size=args.max_embedded_cover_size)
         else:
           play(review, track_urls, merge_with_picture=audio_only)
+        known_reviews.setLastPlayed(review.url)
 
     if track_loop and (args.mode is PlayerMode.MANUAL):
       # update menu and display it
